#include <curses.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <gpm.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>


#include "yascii.h"

global_t globals;



void
main_loop (void)
{
  int my_gpm_fd;
  int ch;
  struct timeval tv;
  fd_set readset;

  tv.tv_sec = 10000;
  tv.tv_usec = 0;
 
  my_gpm_fd = gpm_fd < 0 ? 0 : gpm_fd;
  
  for (;;)
    {

      FD_ZERO (&readset);
      FD_SET (my_gpm_fd, &readset);
      FD_SET (0, &readset);

      if (select (my_gpm_fd + 1, &readset, 0, 0, &tv) < 0)
			continue;

      if (FD_ISSET (my_gpm_fd, &readset) && gpm_fd >= 0) {
		  Gpm_Event evt;

		  if (Gpm_GetEvent (&evt) > 0 && gpm_fd >= 0)
				 globals.mouse_handler.gpm (&evt, &globals.chr);
		  else
				 fprintf (stderr, "Gpm_GetEvent() error\n");
		}


      if (FD_ISSET (0, &readset) && (ch = getch() /*Gpm_Getch()*/ ) != -1)
	{

	  switch (ch)
	    {
		 case KEY_MOUSE:
			 {
			 MEVENT event;
			 getmouse(&event);
			 globals.mouse_handler.curs(&event, &globals.chr);
			 break;
			 }
	    case KEY_F (1):
	      save_file (globals.mainwin);
	      break;
	    case KEY_F (2):
	      load_file (globals.mainwin);
	      break;
	    case KEY_F (5):
	      choose_color (globals.mainwin);
	      break;
	    case KEY_F (6):
	      charmap (globals.mainwin);
	      break;
	    default:
	      globals.chr = ch;
	      break;
	    }
	}

      refresh ();
    }
}

void
main_init (int argc, char **argv)
{
  
  int opt_mask = GPM_DRAG | GPM_DOWN;
  int def_mask = GPM_MOVE | GPM_HARD;
  Gpm_Connect conn;
  int z = 0, bg, fg;

  signal (SIGINT, finish);

  globals.mainwin = NULL;
  globals.chr = 'X';

  globals.mainwin = initscr ();

  keypad (stdscr, 1);
//  nodelay (stdscr, 1);
//  intrflush (stdscr, 1);
	
  noecho ();
  refresh ();

  start_color ();
  use_default_colors ();

  for (bg = 0; bg < 8; bg++)
    for (fg = 0; fg < 8; fg++)
      init_pair (z++, fg, bg);

  assume_default_colors (COLOR_YELLOW, COLOR_BLACK);

  conn.eventMask = opt_mask;
  conn.defaultMask = def_mask;
  conn.maxMod = 0;
  conn.minMod = 0;

  if (Gpm_Open (&conn, 0) == -1) {
      fprintf (stderr, "Can't open mouse connection\n");
      exit (1);
  }
  else if(gpm_fd == -2) {
	  Gpm_Close ();
	  if(mousemask(ALL_MOUSE_EVENTS, (mmask_t *) 0) == 0) beep();
	  globals.mouse_handler.curs = cuser_handler;
  }
  else {
		globals.mouse_handler.gpm = guser_handler;
	   gpm_handler = guser_handler;
  }
}

int
guser_handler (Gpm_Event * event, void *data)
{
  char *buf = data ? data : &globals.chr;

  if (event->margin)
    return 0;

  move (event->y - 1, event->x - 1);

  if (event->buttons == GPM_B_LEFT)
    {
      addch (*buf);
      move (event->y - 1, event->x - 1);
      chgat (1, attrb, color, NULL);
    }
  else if (event->buttons == GPM_B_RIGHT)
    {
      addch (32);
      move (event->y - 1, event->x - 1);
      chgat (1, A_NORMAL, COLOR_NONE, NULL);
    }

  return 0;
}

int cuser_handler(MEVENT* event, void* data)
{
	char* buf = data ? data : &globals.chr;

	move(event->y, event->x);

	if(event->bstate & BUTTON1_CLICKED) {
		addch(*buf);
		move(event->y, event->x);
		chgat(1, attrb, color, NULL);
	}
	else if(event->bstate & BUTTON3_CLICKED) {
		addch(32);
		move(event->y, event->x);
		chgat(1, A_NORMAL, COLOR_NONE, NULL);
	}
	return 0;
}


int
get_real (int z)
{
  int p = (z / 8) % 2;

  if (p)
    z -= 8;

  p = z / 8;
  p = p / 2;

  return z - 8 * p;
}

int
get_foreground (int z)
{
  return z & 7;
}

int
get_background (int z)
{
  return (z >> 3) & 7;
}

int
get_bold (int z)
{
  return (z >> 6) & 1;
}

int
main (int argc, char *argv[])
{

  main_init (argc, argv);
  main_loop ();

  while (Gpm_Close ());
  finish (0);

  exit (0);
}

void
save_file (WINDOW * mainwin)
{
  FILE *file, *ftmp;
  char line[1024];
  int a, b;
  int type = file_type (mainwin);

  if (!(file = file_selection (mainwin, "w")))
    return;
  ftmp = fopen (".tempascii", "w");
  for (a = 0; a < LINES - 5; a++)
    {
      for (b = 0; b < COLS; b++)
	{
	  int ch;
	  char chr;
	  int clr;
	  int atr;
	  int bold = 0;
	  short fg, bg;

	  move (a, b);
	  ch = winch (mainwin);
	  chr = ch & A_CHARTEXT;
	  if (type != TXT)
	    {
	      clr = (ch & A_COLOR) / 256;
	      atr = ch & A_ATTRIBUTES;
	      if (atr & A_BOLD)
		bold = 1;
	      fg = get_foreground (clr);
	      bg = get_background (clr);
	      if (type == IRC)
		{
		  if (bold)
		    fg += 8;	//, bg += 8;
		  if (clr && bg != COLOR_BLACK)
		    fprintf (ftmp, "\003%02d,%02d%c\003", irc_map[fg],
			     irc_map[bg], chr);
		  else if (clr)
		    fprintf (ftmp, "\003%02d%c\003", irc_map[fg], chr);
		  else
		    fprintf (ftmp, "%c", chr);
		}
	      else if (type == ANSI)
		{
		  if (clr && bg != COLOR_BLACK)
		    fprintf (ftmp, "\033[%d;%d;%dm%c\033[0;0m", bold,
			     ansi_map[fg], ansi_map[bg] + 10, chr);
		  else if (clr)
		    fprintf (ftmp, "\033[%d;%d;49m%c\033[0;0m", bold,
			     ansi_map[fg], chr);
		  else
		    fprintf (ftmp, "%c", chr);
		}
	    }
	  else
	    fprintf (ftmp, "%c", ch);
	}
      fprintf (ftmp, "\n");
    }
  fclose (ftmp);
  ftmp = fopen (".tempascii", "r");
  while (fgets (line, 1024, ftmp))
    {
      int i = strlen (line) - 2;

      while (line[i] == 32)
	i--;
      line[i + 1] = 0;
      fputs (line, file);
      fputs ("\n", file);
    }
  fclose (file);
  fclose (ftmp);
  unlink (".tempascii");

}

void
load_file (WINDOW * mainwin)
{
  FILE *file;
  char line[256];
  int x, y = 0;
  int type = file_type (mainwin);

  if (!(file = file_selection (mainwin, "r")))
    return;
  while (fgets (line, 256, file))
    {

      if (type == TXT)
	{
	  for (x = 0; line[x] != 10; x++)
	    {
	      move (y, x);
	      addch (line[x]);
	    }
	  y++;
	}
      else
	{
	  int fg, bg, clr, atr = A_NORMAL;
	  int z;

	  for (x = 0, z = 0; line[x] != 10; x++)
	    {
	      bg = 0;
	      move (y, z++);

	      if (type == IRC && line[x] == 3)
		{
		  if (!isdigit (line[x + 1]))
		    addch (line[++x]);
		  else
		    {
		      fg = (line[x + 1] - 48) * 10 + line[x + 2] - 48;
		      if (fg > 6 && fg != 10 && fg != 14 && fg != 15)
			atr = A_BOLD;
		      else
			atr = A_NORMAL;
		      x += 3;
		      if (line[x] == 44)
			bg = (line[x + 1] - 48) * 10 + line[x + 2] - 48, x +=
			  3;
		      clr =
			bg ? (yairc_map[bg] * 8 +
			      yairc_map[fg]) : yairc_map[fg];
		      addch (line[x]);
		      x++;
		      move (y, z - 1);
		      chgat (1, atr, clr, NULL);
		    }
		}
	      else if (type == ANSI && line[x] == 27)
		{
		  fg = (line[x + 4] - 48) * 10 + line[x + 5] - 48;
		  bg = (line[x + 7] - 48) * 10 + line[x + 8] - 48;
		  clr = bg != 49 ? ((bg - 30 - 10) * 8 + (fg - 30)) : fg - 30;
		  if (line[x + 2] - 48)
		    atr = A_BOLD;
		  else
		    atr = A_NORMAL;
		  x += 10;
		  addch (line[x]);
		  x += 6;
		  move (y, z - 1);
		  chgat (1, atr, clr, NULL);
		}
	    }
	  y++;
	}

    }
  refresh ();
}

FILE *
file_selection (WINDOW * mainwin, const char *mode)
{
  WINDOW *dialog;
  FILE *file;
  char filename[256];
  short winsize_x = COLS - 20, winsize_y = 8;

  dialog =
    newwin (winsize_y, winsize_x, LINES / 2 - (winsize_y / 2),
	    COLS / 2 - (winsize_x / 2));
  wclear (dialog);
  box (dialog, ACS_VLINE, ACS_HLINE);
  mvwprintw (dialog, 2, 2, "Enter a filename:");
  mvwprintw (dialog, 3, 2, "...............................");
  wrefresh (dialog);
  echo ();
  mvwgetnstr (dialog, 3, 2, filename, sizeof (filename));
  noecho ();
  delwin (dialog);
  touchwin (mainwin);
  refresh ();
  if (mode == "w")
    if (!(file = fopen (filename, "r")))
      file = fopen (filename, mode);
    else
      {
	fclose (file);
	if (yesno (mainwin, "File exists. Overwrite?"))
	  file = fopen (filename, mode);
	else
	  file = NULL;
      }
  else if (!(file = fopen (filename, "r")))
    file = NULL;

  return file;
}

void
charmap (WINDOW * mainwin)
{
  WINDOW *dialog;
  int i;
  int n;

  dialog = newwin (24, 70, LINES / 2 - 12, COLS / 2 - 35);	/* FIXME: vive le pifometre */
  wclear (dialog);
  box (dialog, ACS_VLINE, ACS_HLINE);
  mvwprintw (dialog, 3, 15, "choose a char...");

  for (i = 0, n = 0; i < 256; i++)
    {
      if (isprint (i))
	mvwprintw (dialog, 5 + n, i % 32, " %c", i);
      if (i % 32 == 0)
	n++;
    }
  wrefresh (dialog);
}

void
choose_color (WINDOW * mainwin)
{
  WINDOW *dialog;
  int i, j, b, y, ch;
  int chs[3];
  int num;

  dialog = newwin (24, 70, LINES / 2 - 12, COLS / 2 - 35);
  wclear (dialog);
  box (dialog, ACS_VLINE, ACS_HLINE);
  mvwprintw (dialog, 3, 15, "choose your color: [ESC = none]");
  mvwprintw (dialog, 5, 0,
	     "000-001-002-003-004-005-006-007-008-009-010-011-012-013-014-015\n"
	     "016-017-018-019-020-021-022-023-024-025-026-027-028-029-030-031\n"
	     "032-033-034-035-036-037-038-039-040-041-042-043-044-045-046-047\n"
	     "048-049-050-051-052-053-054-055-056-057-058-059-060-061-062-063\n"
	     "064-065-066-067-068-069-070-071-072-073-074-075-076-077-078-079\n"
	     "080-081-082-083-084-085-086-087-088-089-090-091-092-093-094-095\n"
	     "096-097-098-099-100-101-102-103-104-105-106-107-108-109-110-111\n"
	     "112-113-114-115-116-117-118-119-120-121-122-123-124-125-126-127\n");

  i = 0;
  j = 0;
  y = 5;
  b = A_NORMAL;

  for (; j < 64; i += 4, j++)
    {
      mvwchgat (dialog, y, i, 3, b, j, NULL);
      if (b == A_NORMAL && j == 63)
	j = -1, b = A_BOLD, y = 5;
      if (b == A_NORMAL && i == 28)
	i = -4, y++;
      if (b == A_BOLD && i == 60)
	i = 28, y++;
    }

  wrefresh (dialog);

  nodelay (stdscr, 0);
  if ((ch = getch ()) == 27)
    attrb = A_NORMAL, color = COLOR_NONE;
  else
    {
      chs[0] = ch - 48;
      chs[1] = getch () - 48;
      chs[2] = getch () - 48;
      num = chs[0] * 100 + chs[1] * 10 + chs[2];
      if ((num / 8) % 2)
	attrb = A_BOLD;
      else
	attrb = A_NORMAL;
      color = get_real (num);
    }

  nodelay (stdscr, 1);
  delwin (dialog);
  redrawwin (mainwin);
}

int
file_type (WINDOW * mainwin)
{
  WINDOW *dialog;
  short x = 50, y = 8;
  int ch;

  dialog = newwin (y, x, LINES / 2 - (y / 2), COLS / 2 - (x / 2));
  wclear (dialog);
  box (dialog, ACS_VLINE, ACS_HLINE);
  mvwprintw (dialog, 2, 2, "choose the type of the file:");
  mvwprintw (dialog, y - 3, 5, "[t]xt");
  mvwprintw (dialog, y - 3, 12, "[i]rc");
  mvwprintw (dialog, y - 3, 19, "[a]nsi");
  wrefresh (dialog);

  nodelay (stdscr, 0);
  ch = getch ();
  nodelay (stdscr, 1);

  delwin (dialog);
  redrawwin (mainwin);

  return ch;
}

short
yesno (WINDOW * mainwin, char *question)
{
  WINDOW *dialog;
  short x = 50, y = 8, ret;
  int ch;

  dialog = newwin (y, x, LINES / 2 - (y / 2), COLS / 2 - (x / 2));
  wclear (dialog);
  box (dialog, ACS_VLINE, ACS_HLINE);
  mvwprintw (dialog, 2, 2, question);
  mvwprintw (dialog, y - 3, 5, "[Y]es");
  mvwprintw (dialog, y - 3, x - 8, "[N]o");
  wrefresh (dialog);

  nodelay (stdscr, 0);
  ch = getch ();
  nodelay (stdscr, 1);

  if (ch == 89 || ch == 121)
    ret = 1;
  else
    ret = 0;

  delwin (dialog);
  redrawwin (mainwin);

  return ret;
}

static void
finish (int sig)
{
  endwin ();
  exit (0);
}
